package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.B2SolverData
import ru.dbotthepony.kbox2d.api.b2Mul
import ru.dbotthepony.kvector.matrix.ndouble.Matrix2d
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix2d
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times

class MotorJoint(def: MotorJointDef) : AbstractJoint(def) {
	// Solver shared
	var linearOffset: Vector2d = def.linearOffset
		set(value) {
			if (value != field) {
				field = value
				bodyA.isAwake = true
				bodyB.isAwake = true
			}
		}

	var angularOffset: Double = def.angularOffset
		set(value) {
			if (value != field) {
				field = value
				bodyA.isAwake = true
				bodyB.isAwake = true
			}
		}

	private var linearImpulse: Vector2d = Vector2d.ZERO
	private var angularImpulse: Double = 0.0

	var maxForce: Double = def.maxForce
		set(value) {
			require(!value.isNaN()) { "Tried to set NaN force" }
			require(value.isFinite()) { "Tried to set infinite force" }
			require(value >= 0.0) { "Tried to set negative force: $value" }
			field = value
		}

	var maxTorque: Double = def.maxTorque
		set(value) {
			require(!value.isNaN()) { "Tried to set NaN torque" }
			require(value.isFinite()) { "Tried to set infinite torque" }
			require(value >= 0.0) { "Tried to set negative torque: $value" }
			field = value
		}

	var correctionFactor: Double = def.correctionFactor
		set(value) {
			require(!value.isNaN()) { "Tried to set NaN correction factor" }
			require(value.isFinite()) { "Tried to set infinite correction factor" }
			require(value in 0.0 .. 1.0) { "Tried to set correction factor out of bounds: $value" }
			field = value
		}

	// Solver temp
	private var indexA: Int = 0
	private var indexB: Int = 0
	private var rA: Vector2d = Vector2d.ZERO
	private var rB: Vector2d = Vector2d.ZERO
	private var localCenterA: Vector2d = Vector2d.ZERO
	private var localCenterB: Vector2d = Vector2d.ZERO
	private var linearError: Vector2d = Vector2d.ZERO
	private var angularError: Double = 0.0
	private var invMassA: Double = 0.0
	private var invMassB: Double = 0.0
	private var invIA: Double = 0.0
	private var invIB: Double = 0.0
	private var linearMass: MutableMatrix2d = MutableMatrix2d()
	private var angularMass: Double = 0.0

	override fun initVelocityConstraints(data: B2SolverData) {
		this.indexA = this.bodyA.islandIndex
		this.indexB = this.bodyB.islandIndex
		this.localCenterA = this.bodyA.sweep.localCenter
		this.localCenterB = this.bodyB.sweep.localCenter
		this.invMassA = this.bodyA.invMass
		this.invMassB = this.bodyB.invMass
		this.invIA = this.bodyA.invI
		this.invIB = this.bodyB.invI

		val cA = data.positions[this.indexA].c
		val aA = data.positions[this.indexA].a
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w

		val cB = data.positions[this.indexB].c
		val aB = data.positions[this.indexB].a
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		// Compute the effective mass matrix.
		this.rA = b2Mul(qA, this.linearOffset - this.localCenterA)
		this.rB = b2Mul(qB, -this.localCenterB)

		// J = [-I -r1_skew I r2_skew]
		// r_skew = [-ry; rx]

		// Matlab
		// K = [ mA+r1y^2*iA+mB+r2y^2*iB,  -r1y*iA*r1x-r2y*iB*r2x,          -r1y*iA-r2y*iB]
		//     [  -r1y*iA*r1x-r2y*iB*r2x, mA+r1x^2*iA+mB+r2x^2*iB,           r1x*iA+r2x*iB]
		//     [          -r1y*iA-r2y*iB,           r1x*iA+r2x*iB,                   iA+iB]

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		// Upper 2 by 2 of K for point to point
		val K = MutableMatrix2d()
		K.r00 = mA + mB + iA * this.rA.y * this.rA.y + iB * this.rB.y * this.rB.y
		K.r10 = -iA * this.rA.x * this.rA.y - iB * this.rB.x * this.rB.y
		K.r01 = K.r10
		K.r11 = mA + mB + iA * this.rA.x * this.rA.x + iB * this.rB.x * this.rB.x

		this.linearMass = (K.inverse ?: Matrix2d.ZERO).toMutableMatrix2d()

		this.angularMass = iA + iB
		if (this.angularMass > 0.0) {
			this.angularMass = 1.0 / this.angularMass
		}

		this.linearError = cB + this.rB - cA - this.rA
		this.angularError = aB - aA - this.angularOffset

		if (data.step.warmStarting) {
			// Scale impulses to support a variable time step.
			this.linearImpulse *= data.step.dtRatio
			this.angularImpulse *= data.step.dtRatio

			val P = Vector2d(this.linearImpulse.x, this.linearImpulse.y)
			vA -= mA * P
			wA -= iA * (b2Cross(this.rA, P) + this.angularImpulse)
			vB += mB * P
			wB += iB * (b2Cross(this.rB, P) + this.angularImpulse)
		} else {
			this.linearImpulse = Vector2d.ZERO
			this.angularImpulse = 0.0
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solveVelocityConstraints(data: B2SolverData) {
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		val h = data.step.dt
		val inv_h = data.step.inv_dt

		// Solve angular friction
		run {
			val Cdot = wB - wA + inv_h * this.correctionFactor * this.angularError
			var impulse = -this.angularMass * Cdot

			val oldImpulse = this.angularImpulse
			val maxImpulse = h * this.maxTorque
			this.angularImpulse = b2Clamp(this.angularImpulse + impulse, -maxImpulse, maxImpulse)
			impulse = this.angularImpulse - oldImpulse

			wA -= iA * impulse
			wB += iB * impulse
		}

		// Solve linear friction
		run {
			val Cdot = vB + b2Cross(wB, this.rB) - vA - b2Cross(wA, this.rA) + inv_h * this.correctionFactor * this.linearError

			var impulse = -b2Mul(this.linearMass, Cdot)
			val oldImpulse = this.linearImpulse
			this.linearImpulse += impulse

			val maxImpulse = h * this.maxForce

			if (this.linearImpulse.lengthSquared > maxImpulse * maxImpulse) {
				this.linearImpulse = this.linearImpulse.normalized
				this.linearImpulse *= maxImpulse
			}

			impulse = this.linearImpulse - oldImpulse

			vA -= mA * impulse
			wA -= iA * b2Cross(this.rA, impulse)

			vB += mB * impulse
			wB += iB * b2Cross(this.rB, impulse)
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solvePositionConstraints(data: B2SolverData): Boolean {
		return true
	}

	override val anchorA: Vector2d
		get() = bodyA.position
	override val anchorB: Vector2d
		get() = bodyB.position

	override fun getReactionForce(inv_dt: Double): Vector2d {
		return inv_dt * linearImpulse
	}

	override fun getReactionTorque(inv_dt: Double): Double {
		return inv_dt * angularImpulse
	}
}
