package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.client.ClientWorld
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.Color
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import kotlin.math.absoluteValue

enum class Move {
	STAND_STILL,
	MOVE_LEFT,
	MOVE_RIGHT
}

interface IWalkableEntity : IEntity {
	/**
	 * Максимальная скорость передвижения этого AliveMovementController в Starbound Units/секунда
	 *
	 * Скорость передвижения: Это скорость вдоль земли (или в воздухе, если парит) при ходьбе.
	 *
	 * Если вектор скорости вдоль поверхности (или в воздухе, если парит) больше заданного значения,
	 * то сущность быстро тормозит (учитывая силу трения)
	 */
	val topSpeed: Double

	/**
	 * Скорость ускорения сущности в Starbound Units/секунда^2
	 *
	 * Если сущность хочет двигаться вправо или влево,
	 * то она разгоняется с данной скоростью.
	 */
	val moveSpeed: Double

	/**
	 * То, как сущность может влиять на свою скорость в Starbound Units/секунда^2
	 * когда находится в свободном падении или в состоянии невесомости.
	 *
	 * Позволяет в т.ч. игрокам изменять свою траекторию полёта в стиле Quake.
	 */
	val freeFallMoveSpeed: Double

	/**
	 * "Сила", с которой сущность останавливается, если не хочет двигаться.
	 *
	 * Зависит от текущего трения, так как технически является множителем трения поверхности,
	 * на которой стоит сущность.
	 */
	val brakeForce: Double

	/**
	 * Высота шага данной сущности. Данное значение отвечает за то, на сколько блоков
	 * сможет подняться сущность просто двигаясь в одном направлении без необходимости прыгнуть.
	 */
	val jumpForce: Double

	/**
	 * Импульс прыжка данной сущности. Если сущность парит, то данное значение не несёт никакой
	 * полезной нагрузки.
	 */
	val stepSize: Double
}

/**
 * Базовый абстрактный класс, реализующий сущность, которая ходит по земле
 */
abstract class WalkableMovementController<T : IWalkableEntity>(entity: T) : MovementController<T>(entity) {
	init {
		body.isFixedRotation = true
	}

	protected abstract val moveDirection: Move

	var wantsToDuck = false
	open var isDucked = false
		protected set

	override fun thinkPhysics(delta: Double) {
		super.thinkPhysics(delta)
		thinkMovement(delta)
	}

	/**
	 * See [IWalkableEntity.topSpeed]
	 */
	open val topSpeed by entity::topSpeed

	/**
	 * See [IWalkableEntity.moveSpeed]
	 */
	open val moveSpeed by entity::moveSpeed

	/**
	 * See [IWalkableEntity.freeFallMoveSpeed]
	 */
	open val freeFallMoveSpeed by entity::freeFallMoveSpeed

	/**
	 * See [IWalkableEntity.brakeForce]
	 */
	open val brakeForce by entity::brakeForce

	/**
	 * See [IWalkableEntity.stepSize]
	 */
	open val stepSize by entity::stepSize

	/**
	 * See [IWalkableEntity.jumpForce]
	 */
	open val jumpForce by entity::jumpForce

	protected var jumpRequested = false
	protected var nextJump = 0.0

	open fun requestJump() {
		if (jumpRequested || nextJump > world.timer)
			return

		jumpRequested = true
		nextJump = world.timer + 0.1
	}

	open fun recallJump() {
		if (jumpRequested) {
			jumpRequested = false
			nextJump = 0.0
		}
	}

	protected abstract fun canUnDuck(): Boolean

	protected open fun thinkMovement(delta: Double) {
		if (onGround && !isDucked) {
			when (moveDirection) {
				Move.STAND_STILL -> {
					body.linearVelocity += Vector2d(x = -body.linearVelocity.x * delta * brakeForce)
				}

				Move.MOVE_LEFT -> {
					if (body.linearVelocity.x > 0.0) {
						body.linearVelocity += Vector2d(x = -body.linearVelocity.x * delta * brakeForce)
					}

					if (body.linearVelocity.x > -topSpeed) {
						body.linearVelocity += Vector2d(x = -moveSpeed * delta)
					}
				}

				Move.MOVE_RIGHT -> {
					if (body.linearVelocity.x < 0.0) {
						body.linearVelocity += Vector2d(x = -body.linearVelocity.x * delta * brakeForce)
					}

					if (body.linearVelocity.x < topSpeed) {
						body.linearVelocity += Vector2d(x = moveSpeed * delta)
					}
				}
			}

			if (jumpRequested) {
				jumpRequested = false
				nextJump = world.timer + 0.1

				body.linearVelocity += Vector2d(y = jumpForce)
			}
		} else if (!onGround && !isDucked && freeFallMoveSpeed != 0.0) {
			when (moveDirection) {
				Move.STAND_STILL -> {
					// do nothing
				}

				Move.MOVE_LEFT -> {
					body.linearVelocity += Vector2d(x = -freeFallMoveSpeed * delta)
				}

				Move.MOVE_RIGHT -> {
					body.linearVelocity += Vector2d(x = freeFallMoveSpeed * delta)
				}
			}
		} else if (onGround && isDucked) {
			body.linearVelocity += Vector2d(x = -body.linearVelocity.x * delta * brakeForce)
		}

		if (wantsToDuck && onGround) {
			isDucked = true
		} else if (isDucked) {
			if (canUnDuck()) {
				isDucked = false
			}
		}
	}
}

abstract class AliveEntity(world: World<*, *>) : Entity(world) {
	open var maxHealth = 10.0
	open var health = 10.0
}

abstract class AliveWalkingEntity(world: World<*, *>) : AliveEntity(world), IWalkableEntity {
	abstract override val movement: WalkableMovementController<*>

	override val topSpeed = 20.0
	override val moveSpeed = 64.0
	override val freeFallMoveSpeed = 8.0
	override val brakeForce = 16.0
	override val jumpForce = 20.0
	override val stepSize = 1.1

	open var wantsToDuck
		get() = movement.wantsToDuck
		set(value) { movement.wantsToDuck = value }

	open val isDucked get() = movement.isDucked
}
